rm(list = ls())

# load packages.
library(FAdist)
library(np)

# set the seed for random number generation
seed <- 618
set.seed(seed)

h_lb <- 0.0001  # lower bound for optimal bandwidth search
h_ub <- 1  # upper bound for optimal bandwidth search

N <- 250  # sample size: 250, 500, 1000, 2000
numSims <- 5000  # simulation repetitions

beta0 <- matrix(c(1, 0.5, 0.5), nrow = 3, ncol = 1)  # true value for beta

# generate empty matrices to store the estimates
matB_cm_kernel <- matrix(NaN, numSims, 3)  # CM estimates
matB_if <- matrix(NaN, numSims, 3)  # instrumental function estimates
matB_cl <- matrix(NaN, numSims, 3)  # conditional likelihood estimates

# CM functions
## CM criterion function
Qn_cm_kernel <- function(b,normIndex = 1,normVal = 1) {
  if (normIndex == 1)
    beta <- rbind(normVal,b[1],b[2])
  else if (normIndex == 2)
    beta <- rbind(b[1],normVal,b[2])
  else
    beta <- rbind(b[1],b[2],normVal)
  sum(apply(cbind(-deltaXP_kernel %*% beta,0),1,max))
}

## cross validation criterion function
CVQn <- function(h) {
  prob <- matrix(0, N, 6)
  bw <- array(h,dim(X)[2])
  for (j in 1:6) {
    prob[, j] <- npksum(
      X,
      Y[, j],
      leave.one.out = TRUE,
      bandwidth.divide = TRUE,
      bws = bw
    )$ksum /
      npksum(
        X,
        leave.one.out = TRUE,
        bandwidth.divide = TRUE,
        bws = bw
      )$ksum
  }
  sum((Y-prob)^2)
}

# intrumental function method functions
## generate permutations of n objects out of m with replacement
permutation <- function(m,n) {
  if (n==1) {
    return(matrix(1:m,m,1))
  } else {
    sp <- permutation(m,n-1)
    p <- nrow(sp)
    A <- matrix(nrow=p*m,ncol=n)
    for (i in 1:m) {
      A[((i-1)*p+1):(i*p),] <- cbind(matrix(i,p,1),sp)
      
    }
    return(A)
  }
}

gperm <- function(i,ii) {
  deltaXi <- cbind(deltaX[i,1:3], deltaX[i,4:6])
  deltaYi <- deltaY[i,]
  deltaXYi <- deltaXi %*% deltaYi
  indic <- all(X[i,]>=boundmatrix[ii,1:12]) * all(X[i,]<boundmatrix[ii,13:24])
  deltaXYi * indic
}

gnperm <- function(ii) {
  s <- 0
  for (i in 1:N) {
    s <- s + gperm(i,ii)
  }
  s/N
}

## instrumental function method criterion function
Qn_if <- function(b,
                  normIndex = 1,
                  normVal = 1) {
  if (normIndex == 1)
    beta <- rbind(normVal, b[1], b[2])
  else if (normIndex == 2)
    beta <- rbind(b[1], normVal, b[2])
  else
    beta <- rbind(b[1], b[2], normVal)
  
  ss <- 0
  for (ii in 1:nrow(boundmatrix)) {
    ss <- ss + abs(min(t(gnperm(ii)) %*% beta, 0))
  }
  ss / N / (N-1)
}

# conditional likelihood method functions
## conditional likelihood for any two different choices
cll <- function(b, c1, c2) {
  sample1Index <- which(choice[,1] == c1 & choice[,2] == c2)
  if (!is.null(sample1Index)) {
    sample1X <- X[sample1Index,,drop = FALSE]
    sample1X <-
      cbind(matrix(0,length(sample1Index),3),sample1X[,1:6,drop = FALSE],matrix(0,length(sample1Index),3),sample1X[,7:12,drop = FALSE])
    XT2diff <-
      sample1X[,((c2 - 1) * 3 + 1 + 9):(c2 * 3 + 9)] - sample1X[,((c1 - 1) * 3 + 1 + 9):(c1 * 3 + 9)]
    XT1diff <-
      sample1X[,((c2 - 1) * 3 + 1):(c2 * 3)] - sample1X[,((c1 - 1) * 3 + 1):(c1 * 3)]
    prob1 <-
      exp(XT2diff %*% b) / (exp(XT1diff %*% b) + exp(XT2diff %*% b))
    cll1 <- sum(log(prob1)) 
  } else {
    cll1 <- 0
  }
  
  sample2Index <- which(choice[,1] == c2 & choice[,2] == c1)
  if (!is.null(sample2Index)) {
    sample2X <- X[sample2Index,,drop = FALSE]
    if (length(sample2Index) == 1) {
      sample2X <- matrix(sample2X,1,length(sample2X))
    }
    sample2X <-
      cbind(matrix(0,length(sample2Index),3),sample2X[,1:6,drop = FALSE],matrix(0,length(sample2Index),3),sample2X[,7:12,drop = FALSE])
    XT2diff <-
      sample2X[,((c1 - 1) * 3 + 1 + 9):(c1 * 3 + 9)] - sample2X[,((c2 - 1) * 3 + 1 + 9):(c2 * 3 + 9)]
    XT1diff <-
      sample2X[,((c1 - 1) * 3 + 1):(c1 * 3)] - sample2X[,((c2 - 1) * 3 + 1):(c2 * 3)]
    prob2 <-
      exp(XT2diff %*% b) / (exp(XT1diff %*% b) + exp(XT2diff %*% b))
    cll2 <- sum(log(prob2))
  } else {
    cll2 <- 0
  }
  
  (cll1 + cll2) 
  
}

## conditonal likelihood function
ll <- function(b) {
  - cll(b,1,2) - cll(b,1,3) - cll(b,2,3)
}

# main
## generate hypercubes by evenly dividing unit intervals into 5 subintervals suggested in Andrews and Shi (2013)
pair <- matrix(c(0,1/5,1/5,2/5,2/5,3/5,3/5,4/5,4/5,1),5,2,byrow = T)
indexperm <- permutation(5,2)
boundmatrix <- matrix(0,0,24)
tmp1 <- cbind(matrix(0,1,12),matrix(1,1,12))
for (i in 2:12) {
  for (j in 1:(i-1)) {
    for (k in 1:nrow(indexperm)) {
      tmp2 <- tmp1
      tmp2[1,i] <- pair[indexperm[k,1],1]
      tmp2[1,i+12] <- pair[indexperm[k,1],2]
      tmp2[1,j] <- pair[indexperm[k,2],1]
      tmp2[1,j+12] <- pair[indexperm[k,2],2]
      boundmatrix <- rbind(boundmatrix,tmp2)
    }
  }
}

besth <- matrix(NaN, numSims, 1)  # empty matrix to store optimal bandwidth

for (inst in 1:numSims) {
  
  # DGP
  X <- matrix(runif(N * 2 * 3 * 2), nrow = N * 2, ncol = 3 * 2)
  omega <- matrix(runif(N * 2), nrow = N, ncol = 2)
  A <- (cbind(X[(N+1):(2*N),1]-X[1:N,1],X[(N+1):(2*N),4]-X[1:N,4]) + omega) / 4
  
  #u01 <- matrix(rgumbel(N * 2 * 2,scale = 1/6), N * 2, 2) ## uncomment for mlogit design
  u01 <- matrix(rcauchy(N * 2 * 2,0,2),nrow = N * 2,ncol = 2) ## uncomment for Cauchy design
  
  #u2 <- matrix(rgumbel(N * 2,scale = 1/6),N * 2, 1)
  u2 <- matrix(rcauchy(N * 2,0,2),nrow = N * 2,ncol = 1)
  
  u <- cbind(u01, u2)
  eps <- cbind(u[,2] - u[,1],u[,3] - u[,1])
  
  util12 <-
    rbind(A, A) + cbind(X[, 1:3] %*% beta0, X[, 4:6] %*% beta0) + eps
  util <- cbind(matrix(0, N * 2, 1), util12)
  
  choice <- max.col(util)
  Y <- matrix(0, N * 2, 3)
  for (itr in 1:(N * 2)) {
    Y[itr, choice[itr]] <- 1
  }
  X <- cbind(X[1:N, ], X[(N + 1):(N * 2), ])
  Y <- cbind(Y[1:N, ], Y[(N + 1):(N * 2), ])
  deltaX <- X[, 1:6] - X[, 7:12]
  deltaY <- Y[, 2:3] - Y[, 5:6]
  choice <- cbind(choice[1:N], choice[(N + 1):(N * 2)])
  
  # compute conditional probabilities 
  prob <- matrix(0, N, 6)
  h <- optimize(CVQn,c(h_lb,h_ub))  ## CV for optimal h
  besth[inst,] <- h$minimum
  bw <- array(h$minimum,dim(X)[2])
  for (j in 1:6) {
    prob[, j] <- npksum(
      X,
      Y[, j],
      leave.one.out = TRUE,
      bandwidth.divide = TRUE,
      bws = bw
    )$ksum /
      npksum(
        X,
        leave.one.out = TRUE,
        bandwidth.divide = TRUE,
        bws = bw
      )$ksum
  }
  cprob_kernel <- prob
  deltaP_kernel <- cprob_kernel[, 2:3] - cprob_kernel[, 5:6]
  deltaXP_kernel <- deltaX[,1:3] * matrix(rep(deltaP_kernel[,1],3),N,3) + deltaX[,4:6] * matrix(rep(deltaP_kernel[,2],3),N,3)
  
  # empty values to store beta estimates and optimization values
  optObj_cm_kernel <- Inf
  optObj_if <- Inf
  optObj_cl <- Inf
  optBeta_cm_kernel <- NaN
  optBeta_if <- NaN
  optBeta_cl <- NaN
  
  # do optimizations
  ui <- matrix(c(1,0,-1,0,0,1,0,-1),4,2)
  ci <- matrix(-1,4,1)
  opt_cm_kernel <- constrOptim(c(0, 0), Qn_cm_kernel,NULL,ui,ci)
  opt_if <- optim(c(0, 0), Qn_if)
  opt_cl <- optim(c(1,0,0),ll)
  
  optBeta_cm_kernel <- rbind(1, opt_cm_kernel$par[1], opt_cm_kernel$par[2])
  optBeta_if <- rbind(1, opt_if$par[1], opt_if$par[2])
  optBeta_cl <- opt_cl$par
  
  matB_cm_kernel[inst,] <- optBeta_cm_kernel
  matB_if[inst,] <- optBeta_if
  matB_cl[inst,] <- optBeta_cl / optBeta_cl[1]
}

# output estimates
write.table(besth,
            file = "besth",
            row.names = FALSE,
            col.names = FALSE)
write.table(matB_cm_kernel,
            file = "matB_cm_kernel",
            row.names = FALSE,
            col.names = FALSE)
write.table(matB_if,
            file = "matB_if",
            row.names = FALSE,
            col.names = FALSE)
write.table(matB_cl,
            file = "matB_cl",
            row.names = FALSE,
            col.names = FALSE)
